"""
Test for the expansion of packages.
"""

import pytest  # type:ignore[import]
from lark import Token, Tree

from ahbicht.expressions.expression_resolver import parse_expression_including_unresolved_subexpressions


class TestTimeConditionReplacement:
    """
    Test for the replacement of time conditions
    """

    @pytest.mark.parametrize(
        "expression, replace_time_conditions, expected_tree",
        [
            pytest.param(
                "[UB1] U [42]",
                False,
                Tree(  # type:ignore[misc]
                    "and_composition",
                    [
                        Tree(Token("RULE", "time_condition"), [Token("TIME_CONDITION_KEY", "UB1")]),
                        Tree(Token("RULE", "condition"), [Token("CONDITION_KEY", "42")]),
                    ],
                ),
            ),
            pytest.param(
                "[UB1] U [42]",
                True,
                Tree(  # type:ignore[misc]
                    "and_composition",
                    [
                        Tree(Token("RULE", "condition"), [Token("CONDITION_KEY", "932")]),
                        Tree(Token("RULE", "condition"), [Token("CONDITION_KEY", "42")]),
                    ],
                ),
            ),
            pytest.param(
                "[UB2] U [42]",
                False,
                Tree(  # type:ignore[misc]
                    "and_composition",
                    [
                        Tree(Token("RULE", "time_condition"), [Token("TIME_CONDITION_KEY", "UB2")]),
                        Tree(Token("RULE", "condition"), [Token("CONDITION_KEY", "42")]),
                    ],
                ),
            ),
            pytest.param(
                "[UB2] U [42]",
                True,
                Tree(  # type:ignore[misc]
                    "and_composition",
                    [
                        Tree(Token("RULE", "condition"), [Token("CONDITION_KEY", "934")]),
                        Tree(Token("RULE", "condition"), [Token("CONDITION_KEY", "42")]),
                    ],
                ),
            ),
            pytest.param(
                "[UB3] U [42]",
                False,
                Tree(  # type:ignore[misc]
                    "and_composition",
                    [
                        Tree(Token("RULE", "time_condition"), [Token("TIME_CONDITION_KEY", "UB3")]),
                        Tree(Token("RULE", "condition"), [Token("CONDITION_KEY", "42")]),
                    ],
                ),
            ),
            pytest.param(
                "[UB3] U [42]",
                True,
                Tree(  # type:ignore[misc]
                    "and_composition",
                    [
                        Tree(
                            "xor_composition",
                            [
                                Tree(
                                    "then_also_composition",
                                    [
                                        Tree(Token("RULE", "condition"), [Token("CONDITION_KEY", "932")]),
                                        Tree(Token("RULE", "condition"), [Token("CONDITION_KEY", "492")]),
                                    ],
                                ),
                                Tree(
                                    "then_also_composition",
                                    [
                                        Tree(Token("RULE", "condition"), [Token("CONDITION_KEY", "934")]),
                                        Tree(Token("RULE", "condition"), [Token("CONDITION_KEY", "493")]),
                                    ],
                                ),
                            ],
                        ),
                        Tree(Token("RULE", "condition"), [Token("CONDITION_KEY", "42")]),
                    ],
                ),
            ),
        ],
    )
    async def test_time_condition_expansion(
        self, expression: str, replace_time_conditions: bool, expected_tree: Tree[Token]
    ):
        parsed_tree = await parse_expression_including_unresolved_subexpressions(
            expression, replace_time_conditions=replace_time_conditions
        )
        assert parsed_tree == expected_tree



"""
functions related to content evaluation
"""
import asyncio
from typing import Any, Awaitable, Callable, List, Optional, Tuple, Union

from lark import Token, Tree

from ahbicht.content_evaluation.content_evaluation_result import ContentEvaluationResult
from ahbicht.content_evaluation.evaluationdatatypes import EvaluatableData
from ahbicht.content_evaluation.fc_evaluators import FcEvaluator
from ahbicht.content_evaluation.rc_evaluators import ContentEvaluationResultBasedRcEvaluator, RcEvaluator
from ahbicht.content_evaluation.token_logic_provider import TokenLogicProvider
from ahbicht.expressions import InvalidExpressionError
from ahbicht.expressions.ahb_expression_evaluation import evaluate_ahb_expression_tree
from ahbicht.expressions.condition_expression_parser import extract_categorized_keys_from_tree
from ahbicht.expressions.expression_resolver import parse_expression_including_unresolved_subexpressions


async def is_valid_expression(
    expression_or_tree: Union[str, Tree[Token]],
    content_evaluation_result_setter: Callable[[ContentEvaluationResult], Any],
) -> Tuple[bool, Optional[str]]:
    """
    Returns true iff the given expression is both well-formed and valid.
    An expression is valid if and only if all possible content evaluations lead to a meaningful results.
    ⚠ This only works if the provided content_evaluation_result_setter writes the EvaluatableData in such a way, that
    the injected Evaluators (FC/RC/Hints/Package) can work with it.
    This is easiest for the ContentEvaluationResultBased FC/RC/Hints/Package token logic providers.
    :param content_evaluation_result_setter: a threadsafe method that writes the given Content Evaluation Result into
    the evaluatable data
    :param expression_or_tree: "Muss [1] U [2]" (returns True)  "Muss ([61]u [584]) o[583]" (returns False)
    :return: (True,None) iff the expression is valid; (False, error message) otherwise
    """
    tree: Tree[Token]
    if isinstance(expression_or_tree, str):
        try:
            tree = await parse_expression_including_unresolved_subexpressions(expression_or_tree)
        except SyntaxError as syntax_error:
            return False, str(syntax_error)
    elif isinstance(expression_or_tree, Tree):
        tree = expression_or_tree
    else:
        raise ValueError(f"{expression_or_tree} is neither a string nor a Tree")
    categorized_key_extract = extract_categorized_keys_from_tree(tree, sanitize=True)
    evaluation_tasks: List[Awaitable] = []
    for content_evaluation_result in categorized_key_extract.generate_possible_content_evaluation_results():
        # create (but do not await) the evaluation tasks for all possible content evaluation results
        # the idea is, that if _any_ evaluation task raises an uncatched exception this can be interpreted as:
        # "the expression is invalid"
        async def evaluate_with_cer(cer: ContentEvaluationResult):
            content_evaluation_result_setter(cer)
            try:
                await evaluate_ahb_expression_tree(tree)
            except NotImplementedError as not_implemented_error:
                # we can ignore some specific errors
                if "due to missing information" in str(not_implemented_error):
                    pass  # this happens for UNKNOWN; it's okay because the expression might still be valid
                else:
                    raise not_implemented_error  # this is, in general, an indicator for an invalid expression
            except InvalidExpressionError as single_invalid_expression_error:
                if isinstance(expression_or_tree, str):
                    single_invalid_expression_error.invalid_expression = expression_or_tree
                raise

        evaluation_tasks.append(evaluate_with_cer(content_evaluation_result))
    try:
        # await the previously collected (yet unawaited) tasks
        await asyncio.gather(*evaluation_tasks)
    except InvalidExpressionError as invalid_expression_error:
        # if any evaluation throws a InvalidExpressionError the expression is invalid
        return False, invalid_expression_error.error_message
    return True, None



""" Tests the HintsProvider, that assumes a ContentEvaluationResult to be present in the evaluatable data"""
from typing import Optional

import inject
import pytest  # type:ignore[import]
from _pytest.fixtures import SubRequest  # type:ignore[import]

from ahbicht.content_evaluation.content_evaluation_result import ContentEvaluationResult
from ahbicht.content_evaluation.token_logic_provider import TokenLogicProvider
from unittests.defaults import default_test_format, default_test_version


class TestCerBasedHintsProvider:
    """Test for the evaluation using the ContentEvaluationResult Based Hints Provider"""

    @pytest.mark.parametrize(
        "inject_cer_evaluators",
        [
            pytest.param(
                ContentEvaluationResult(
                    format_constraints={},
                    requirement_constraints={},
                    hints={"501": "Hello", "502": "World"},
                )
            )
        ],
        indirect=True,
    )
    @pytest.mark.parametrize(
        "condition_key, expected_result",
        [
            pytest.param("501", "Hello"),
            pytest.param("502", "World"),
            pytest.param("503", None),
        ],
    )
    async def test_evaluation(self, condition_key: str, expected_result: Optional[str], inject_cer_evaluators):
        token_logic_provider: TokenLogicProvider = inject.instance(TokenLogicProvider)  # type:ignore[assignment]
        hints_provider = token_logic_provider.get_hints_provider(default_test_format, default_test_version)
        actual = await hints_provider.get_hint_text(condition_key)
        assert actual == expected_result



# This file makes the unittests directory a package in order to allow the contained test modules to import each other.
# Some IDEs handle this automatically, however a pure `$ pytest unittest` call requires this file.
#
# See also https://docs.pytest.org/en/6.2.x/goodpractices.html



from contextvars import ContextVar
from typing import Optional

import inject
import pytest  # type:ignore[import]

from ahbicht.content_evaluation import is_valid_expression
from ahbicht.content_evaluation.content_evaluation_result import ContentEvaluationResult, ContentEvaluationResultSchema
from ahbicht.content_evaluation.evaluationdatatypes import EvaluatableData, EvaluatableDataProvider
from ahbicht.content_evaluation.evaluator_factory import create_content_evaluation_result_based_evaluators
from ahbicht.content_evaluation.token_logic_provider import SingletonTokenLogicProvider, TokenLogicProvider
from ahbicht.expressions.ahb_expression_parser import parse_ahb_expression_to_single_requirement_indicator_expressions
from ahbicht.expressions.expression_resolver import parse_expression_including_unresolved_subexpressions
from unittests.defaults import default_test_format, default_test_version

_content_evaluation_result: ContextVar[Optional[ContentEvaluationResult]] = ContextVar(
    "_content_evaluation_result", default=None
)


def _get_evaluatable_data():
    """
    returns the _content_evaluation_result context var value wrapped in a EvaluatableData container.
    This is the kind of data that the ContentEvaluationResultBased RC/FC Evaluators, HintsProvider and Package Resolver
    require.
    :return:
    """
    cer = _content_evaluation_result.get()
    return EvaluatableData(
        body=ContentEvaluationResultSchema().dump(cer),
        edifact_format=default_test_format,
        edifact_format_version=default_test_version,
    )


class TestValidityCheck:
    """
    a test class for the expression validation feature
    """

    @pytest.fixture
    def inject_cer_evaluators(self):
        def configure(binder):
            binder.bind(
                TokenLogicProvider,
                SingletonTokenLogicProvider(
                    [*create_content_evaluation_result_based_evaluators(default_test_format, default_test_version)]
                ),
            )
            binder.bind_to_provider(EvaluatableDataProvider, _get_evaluatable_data)

        inject.configure_once(configure)
        yield
        inject.clear()

    @pytest.mark.parametrize(
        "ahb_expression,expected_result",
        [
            pytest.param("Foo", False),
            pytest.param("Muss [1] U [2]", True),
            pytest.param("Muss [61] O [584]", False),  # connecting a hint with LOR is not valid
            pytest.param("Muss [123] O [584]", False),
            pytest.param("Muss [501] X [999]", False),  # connecting a hint XOR fc is not valid
            pytest.param("Muss [501] O [999]", False),  # connecting a hint LOR fc is not valid
            pytest.param("Muss [983][1] X [984][2]", True),
        ],
    )
    async def test_is_valid_expression(self, ahb_expression: str, expected_result: bool, inject_cer_evaluators):
        actual_str = await is_valid_expression(ahb_expression, lambda cer: _content_evaluation_result.set(cer))
        assert actual_str[0] == expected_result
        # check the tree as argument, too
        try:
            tree = await parse_expression_including_unresolved_subexpressions(ahb_expression)
        except SyntaxError:
            return  # ok, the syntax error is actually raised on parsing already
        actual_tree = await is_valid_expression(tree, lambda cer: _content_evaluation_result.set(cer))
        assert actual_tree[0] == expected_result

    async def test_is_valid_expression_value_error(self):
        with pytest.raises(ValueError):
            await is_valid_expression(12345, None)



"""
Functions that are not clearly related to another module
"""
import asyncio
import inspect
from typing import Awaitable, Callable, List, TypeVar, Union

from lark import Tree

from ahbicht.expressions import parsing_logger

Result = TypeVar("Result")


async def gather_if_necessary(results_and_awaitable_results: List[Union[Result, Awaitable[Result]]]) -> List[Result]:
    """
    Await the awaitables, pass the un-awaitable results
    :param results_and_awaitable_results: heterogenous list of both Ts and Awaitable[T]s.
    :return: list of T in the same order as in the input param.
    """
    awaitable_indexes = [n for n, x in enumerate(results_and_awaitable_results) if inspect.isawaitable(x)]
    awaited_results = await asyncio.gather(*[x for x in results_and_awaitable_results if inspect.isawaitable(x)])
    result: List[Result] = []
    awaited_results_index = 0
    for index, obj in enumerate(results_and_awaitable_results):
        if index in awaitable_indexes:
            result.append(awaited_results[awaited_results_index])
            awaited_results_index += 1
        else:
            # we are sure obj is of type T
            result.append(obj)  # type:ignore[arg-type]
    return result


_CACHE_LOG_LEVEL = 5
"""
This is a custom log level, which is smaller than debug: https://docs.python.org/3/library/logging.html#logging-levels
We use it to log cache accesses but not spam at log level DEBUG.
"""


def tree_copy(lru_cached_parsing_func: Callable[[str], Tree]):
    """
    A decorator that returns copy of the cached result from the lru_cached_parsing_func.
    Rationale: We want to cache the tree for various expressions because this is definitely faster than re-parsing it.
    But we don't want the same instance of the tree to be returned over and over again, because the calling code might
    modify the tree and (if we always returned the same instance) might also modify the cache entry. We don't want this.
    The tree_copy decorator is used together shall be used with a @lru_cached function. It returns a copy of the cached
    value instead of the same instance.
    :param lru_cached_parsing_func: A function that parses a string to a tree which is decorated with @lru_cache.
    :return: the decorated function that always returns a copy of the cached result instead of the same instance
    """

    def decorated(*args, **kwargs):
        cache_size_before_parsing = lru_cached_parsing_func.cache_info().currsize
        tree_result: Tree = lru_cached_parsing_func(*args, **kwargs)
        cache_size_after_parsing = lru_cached_parsing_func.cache_info().currsize
        if cache_size_after_parsing == cache_size_before_parsing:
            parsing_logger.log(_CACHE_LOG_LEVEL, "The parsed tree for '%s' has been loaded from the cache", args[0])
        return tree_result.copy()

    return decorated
